"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Final, Iterator, List, Optional, Sequence, Tuple, Union

from .enums import HubType, try_enum
from .metadata import Metadata
from .utils import SequenceProxy, SnowflakeList, cached_slot_property, murmurhash32, utcnow

if TYPE_CHECKING:
    from .abc import Snowflake
    from .guild import Guild
    from .state import ConnectionState
    from .types.experiment import (
        Filters as FiltersPayload,
        GuildExperiment as GuildExperimentPayload,
        Override as OverridePayload,
        Population as PopulationPayload,
        Rollout as RolloutPayload,
        UserExperiment as AssignmentPayload,
    )

__all__ = (
    'ExperimentRollout',
    'ExperimentFilters',
    'ExperimentPopulation',
    'ExperimentOverride',
    'HoldoutExperiment',
    'GuildExperiment',
    'UserExperiment',
)


class ExperimentRollout:
    """Represents a rollout for an experiment population.

    .. container:: operations

        .. describe:: x in y

            Checks if a position is eligible for the rollout.

    .. versionadded:: 2.1

    Attributes
    -----------
    population: :class:`ExperimentPopulation`
        The population this rollout belongs to.
    bucket: :class:`int`
        The bucket the rollout grants.
    ranges: List[Tuple[:class:`int`, :class:`int`]]
        The position ranges of the rollout.
    """

    __slots__ = ('population', 'bucket', 'ranges')

    def __init__(self, population: ExperimentPopulation, data: RolloutPayload):
        bucket, ranges = data

        self.population = population
        self.bucket: int = bucket
        self.ranges: List[Tuple[int, int]] = [(range['s'], range['e']) for range in ranges]

    def __repr__(self) -> str:
        return f'<ExperimentRollout bucket={self.bucket} ranges={self.ranges!r}>'

    def __contains__(self, item: int, /) -> bool:
        for start, end in self.ranges:
            if start <= item <= end:
                return True
        return False


class ExperimentFilters:
    """Represents a number of filters for an experiment population.
    A guild must fulfill all filters to be eligible for the population.

    This is a purposefuly very low-level object.

    .. container:: operations

        .. describe:: x in y

            Checks if a guild fulfills the filter requirements.

    .. versionadded:: 2.1

    Attributes
    -----------
    population: :class:`ExperimentPopulation`
        The population this filter belongs to.
    options: :class:`Metadata`
        The parameters for the filter. If known, murmur3-hashed keys are unhashed to their original names.

        .. note::

            You should query parameters via the properties rather than using this directly.
    """

    __slots__ = ('population', 'options')

    # Most of these are taken from the client
    FILTER_KEYS: Final[Dict[int, str]] = {
        1604612045: 'guild_has_feature',
        2404720969: 'guild_id_range',
        3730341874: 'guild_age_range_days',
        2918402255: 'guild_member_count_range',
        3013771838: 'guild_ids',
        4148745523: 'guild_hub_types',
        188952590: 'guild_has_vanity_url',
        2294888943: 'guild_in_range_by_hash',
        3399957344: 'min_id',
        1238858341: 'max_id',
        2690752156: 'hash_key',
        1982804121: 'target',
        1183251248: 'guild_features',
    }

    def __init__(self, population: ExperimentPopulation, data: FiltersPayload):
        self.population = population
        self.options: Metadata = self.array_object(data)

    def __repr__(self) -> str:
        keys = (
            'features',
            'id_range',
            'age_range',
            'member_count_range',
            'ids',
            'hub_types',
            'range_by_hash',
            'has_vanity_url',
        )
        attrs = [f'{attr}={getattr(self, attr)!r}' for attr in keys if getattr(self, attr) is not None]
        if attrs:
            return f'<ExperimentFilters {" ".join(attrs)}>'
        return '<ExperimentFilters>'

    def __contains__(self, guild: Guild, /) -> bool:
        return self.is_eligible(guild)

    @classmethod
    def array_object(cls, array: list) -> Metadata:
        metadata = Metadata()
        for key, value in array:
            try:
                key = cls.FILTER_KEYS[int(key)]
            except (KeyError, ValueError):
                pass
            if isinstance(value, str) and value.isdigit():
                value = int(value)
            elif value and isinstance(value, list) and isinstance(value[0], list):
                value = cls.array_object(value)
            metadata[str(key)] = value
        return metadata

    @staticmethod
    def in_range(num: float, start: Optional[float], end: Optional[float], /) -> bool:
        if start is not None and num < start:
            return False
        if end is not None and num > end:
            return False
        return True

    @property
    def features(self) -> Optional[List[str]]:
        """Optional[List[:class:`str`]]: The guild features that are eligible for the population."""
        features_filter = self.options.guild_has_feature
        if features_filter is not None:
            return features_filter.guild_features

    @property
    def id_range(self) -> Optional[Tuple[Optional[int], Optional[int]]]:
        """Optional[Tuple[Optional[:class:`int`], Optional[:class:`int`]]]: The range of guild IDs that are eligible for the population."""
        id_range_filter = self.options.guild_id_range
        if id_range_filter is not None:
            return id_range_filter.min_id, id_range_filter.max_id

    @property
    def age_range(self) -> Optional[Tuple[Optional[int], Optional[int]]]:
        """Optional[Tuple[Optional[:class:`int`], Optional[:class:`int`]]]: The range of guild ages that are eligible for the population."""
        age_range_filter = self.options.guild_age_range_days
        if age_range_filter is not None:
            return age_range_filter.min_id, age_range_filter.max_id

    @property
    def member_count_range(self) -> Optional[Tuple[Optional[int], Optional[int]]]:
        """Optional[Tuple[Optional[:class:`int`], Optional[:class:`int`]]]: The range of guild member counts that are eligible for the population."""
        member_count_range_filter = self.options.guild_member_count_range
        if member_count_range_filter is not None:
            return member_count_range_filter.min_id, member_count_range_filter.max_id

    @property
    def ids(self) -> Optional[List[int]]:
        """Optional[List[:class:`int`]]: The guild IDs that are eligible for the population."""
        ids_filter = self.options.guild_ids
        if ids_filter is not None:
            return ids_filter.guild_ids

    @property
    def hub_types(self) -> Optional[List[HubType]]:
        """Optional[List[:class:`HubType`]]: The Student Hub types that are eligible for the population."""
        hub_types_filter = self.options.guild_hub_types
        if hub_types_filter is not None:
            return [try_enum(HubType, hub_type) for hub_type in hub_types_filter.guild_hub_types]

    @property
    def range_by_hash(self) -> Optional[Tuple[int, int]]:
        """Optional[Tuple[:class:`int`, :class:`int`]]: The special rollout position limits on the population."""
        range_by_hash_filter = self.options.guild_in_range_by_hash
        if range_by_hash_filter is not None:
            return range_by_hash_filter.hash_key, range_by_hash_filter.target

    @property
    def has_vanity_url(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Whether a vanity is or is not required to be eligible for the population."""
        has_vanity_url_filter = self.options.guild_has_vanity_url
        if has_vanity_url_filter is not None:
            return has_vanity_url_filter.target

    def is_eligible(self, guild: Guild, /) -> bool:
        """Checks whether the guild fulfills the filter requirements.

        .. note::

            This function is not intended to be used directly. Instead, use :func:`GuildExperiment.bucket_for`.

        Parameters
        -----------
        guild: :class:`Guild`
            The guild to check.

        Returns
        --------
        :class:`bool`
            Whether the guild fulfills the filter requirements.
        """
        features = self.features
        if features is not None:
            # At least one feature must be present
            if not any(feature in guild.features for feature in features):
                return False

        id_range = self.id_range
        if id_range is not None:
            # Guild must be within the range of snowflakes
            if not self.in_range(guild.id, *id_range):
                return False

        age_range = self.age_range
        if age_range is not None:
            # Guild must be within the range of ages (in days)
            if not self.in_range((utcnow() - guild.created_at).days, *age_range):
                return False

        member_count_range = self.member_count_range
        if member_count_range is not None and guild.member_count is not None:
            # Guild must be within the range of member counts
            if not self.in_range(guild.member_count, *member_count_range):
                return False

        ids = self.ids
        if ids is not None:
            # Guild must be in the list of snowflakes, similar to ExperimentOverride
            if guild.id not in ids:
                return False

        hub_types = self.hub_types
        if hub_types is not None:
            # Guild must be a hub in the list of hub types
            if not guild.hub_type or guild.hub_type not in hub_types:
                return False

        range_by_hash = self.range_by_hash
        if range_by_hash is not None:
            # Guild must fulfill the additional population requirements
            hash_key, target = range_by_hash
            result = murmurhash32(f'{hash_key}:{guild.id}', signed=False)
            if result > 0:
                result += result
            else:
                result = (result % 0x100000000) >> 0
            if target and (result % 10000) >= target:
                return False

        has_vanity_url = self.has_vanity_url
        if has_vanity_url is not None:
            # Guild must or must not have a vanity URL
            if not bool(guild.vanity_url_code) == has_vanity_url:
                return False

        return True


class ExperimentPopulation:
    """Represents a population of an experiment.

    .. container:: operations

        .. describe:: x in y

            Checks if a guild is present in the population.

    .. versionadded:: 2.1

    Attributes
    -----------
    experiment: :class:`GuildExperiment`
        The experiment this population belongs to.
    filters: :class:`ExperimentFilters`
        The filters that apply to the population.
    rollouts: List[Tuple[:class:`int`, :class:`int`]]
        The position-based rollouts of the population.
    """

    __slots__ = ('experiment', 'filters', 'rollouts')

    def __init__(self, experiment: GuildExperiment, data: PopulationPayload):
        rollouts, filters = data

        self.experiment = experiment
        self.filters: ExperimentFilters = ExperimentFilters(self, filters)
        self.rollouts: List[ExperimentRollout] = [ExperimentRollout(self, x) for x in rollouts]

    def __repr__(self) -> str:
        return f'<ExperimentPopulation experiment={self.experiment!r} filters={self.filters!r} rollouts={self.rollouts!r}>'

    def __contains__(self, item: Guild, /) -> bool:
        return self.bucket_for(item) != -1

    def is_eligible(self, guild: Guild, /) -> bool:
        """Checks whether the guild is eligible for the population.

        .. note::

            This function is not intended to be used directly. Instead, use :func:`GuildExperiment.bucket_for`.

        Parameters
        -----------
        guild: :class:`Guild`
            The guild to check.

        Returns
        --------
        :class:`bool`
            Whether the guild is eligible for the population.
        """
        return self.filters.is_eligible(guild)

    def bucket_for(self, guild: Guild, _result: Optional[int] = None, /) -> int:
        """Returns the assigned experiment bucket within a population for a guild.
        Defaults to None (-1) if the guild is not in the population.

        .. note::

            This function is not intended to be used directly. Instead, use :func:`GuildExperiment.bucket_for`.

        Parameters
        -----------
        guild: :class:`Guild`
            The guild to compute experiment eligibility for.

        Raises
        ------
        :exc:`ValueError`
            The experiment name is unset.

        Returns
        -------
        :class:`int`
            The experiment bucket.
        """
        if not self.filters.is_eligible(guild):
            return -1

        if _result is None:
            _result = self.experiment.result_for(guild)

        for rollout in self.rollouts:
            for start, end in rollout.ranges:
                if start <= _result <= end:
                    return rollout.bucket

        return -1


class ExperimentOverride:
    """Represents an experiment override.

    .. container:: operations

        .. describe:: len(x)

            Returns the number of resources eligible for the override.

        .. describe:: x in y

            Checks if a resource is eligible for the override.

        .. describe:: iter(x)

            Returns an iterator of the resources eligible for the override.

    .. versionadded:: 2.1

    Attributes
    -----------
    experiment: :class:`GuildExperiment`
        The experiment this override belongs to.
    bucket: :class:`int`
        The bucket the override applies.
    """

    __slots__ = ('experiment', 'bucket', '_ids')

    def __init__(self, experiment: GuildExperiment, data: OverridePayload):
        self.experiment = experiment
        self.bucket: int = data['b']
        self._ids: SnowflakeList = SnowflakeList(map(int, data['k']))

    def __repr__(self) -> str:
        return f'<ExperimentOverride bucket={self.bucket} ids={self.ids!r}>'

    def __len__(self) -> int:
        return len(self._ids)

    def __contains__(self, item: Union[int, Snowflake], /) -> bool:
        if isinstance(item, int):
            return item in self._ids
        return item.id in self._ids

    def __iter__(self) -> Iterator[int]:
        return iter(self._ids)

    @property
    def ids(self) -> Sequence[int]:
        """Sequence[:class:`int`]: The eligible guild IDs for the override."""
        return SequenceProxy(self._ids)


class HoldoutExperiment:
    """Represents an experiment holdout.

    .. container:: operations

        .. describe:: x in y

            Checks if a guild fulfills the dependency.

    .. versionadded:: 2.1

    Attributes
    -----------
    dependent: :class:`GuildExperiment`
        The experiment that this holdout blocks.
    name: :class:`str`
        The name of the dependency.
    bucket: :class:`int`
        The required bucket of the dependency.
    """

    __slots__ = ('dependent', 'name', 'bucket', '_hash')

    def __init__(self, dependent: GuildExperiment, name: str, bucket: int):
        self.dependent = dependent
        self.name: str = name
        self.bucket: int = bucket

    def __repr__(self) -> str:
        return f'<HoldoutExperiment name={self.name!r} bucket={self.bucket}>'

    def __contains__(self, item: Guild) -> bool:
        return self.is_blocked(item)

    @cached_slot_property('_hash')
    def hash(self) -> int:
        """:class:`int`: The 32-bit unsigned Murmur3 hash of the dependency name."""
        return murmurhash32(self.name, signed=False)

    @property
    def experiment(self) -> Optional[GuildExperiment]:
        """Optional[:class:`GuildExperiment`]: The experiment dependency, if found."""
        experiment = self.dependent._state.guild_experiments.get(self.hash)
        if experiment and not experiment.name:
            # Backfill the name
            experiment._name = self.name
        return experiment

    def is_blocked(self, guild: Guild, /) -> bool:
        """Checks whether the guild is blocked by the holdout.

        .. note::

            This function is not intended to be used directly. Instead, use :func:`GuildExperiment.bucket_for`.

        Parameters
        -----------
        guild: :class:`Guild`
            The guild to check.

        Returns
        --------
        :class:`bool`
            Whether the guild is blocked by the holdout.
        """
        experiment = self.experiment
        if self.hash == self.dependent.hash or experiment is None:
            # We don't have the experiment so there is no holdout
            return False

        return experiment.bucket_for(guild) == self.bucket


class GuildExperiment:
    """Represents a guild experiment rollout.

    .. container:: operations

        .. describe:: x == y

            Checks if two experiments are equal.

        .. describe:: x != y

            Checks if two experiments are not equal.

        .. describe:: hash(x)

            Returns the experiment's hash.

    .. versionadded:: 2.1

    Attributes
    -----------
    hash: :class:`int`
        The 32-bit unsigned Murmur3 hash of the experiment's name.
    revision: :class:`int`
        The current revision of the experiment rollout.
    populations: List[:class:`ExperimentPopulation`]
        The rollout populations of the experiment.
    overrides: List[:class:`ExperimentOverride`]
        The explicit bucket overrides of the experiment.
    overrides_formatted: List[List[:class:`ExperimentPopulation`]]
        Additional rollout populations for the experiment.
    holdout: Optional[:class:`HoldoutExperiment`]
        An experiment that blocks the rollout of this experiment.
    aa_mode: :class:`bool`
        Whether the experiment is in A/A mode.
    trigger_debugging: :class:`bool`
        Whether experiment analytics trigger debugging is enabled.
    """

    __slots__ = (
        '_state',
        'hash',
        '_name',
        'revision',
        'populations',
        'overrides',
        'overrides_formatted',
        'holdout',
        'aa_mode',
        'trigger_debugging',
    )

    def __init__(self, *, state: ConnectionState, data: GuildExperimentPayload):
        (
            hash,
            hash_key,
            revision,
            populations,
            overrides,
            overrides_formatted,
            holdout_name,
            holdout_bucket,
            aa_mode,
            trigger_debugging,
            *_,
        ) = data

        self._state = state
        self.hash: int = hash
        self._name: Optional[str] = hash_key
        self.revision: int = revision
        self.populations: List[ExperimentPopulation] = [ExperimentPopulation(self, x) for x in populations]
        self.overrides: List[ExperimentOverride] = [ExperimentOverride(self, x) for x in overrides]
        self.overrides_formatted: List[List[ExperimentPopulation]] = [
            [ExperimentPopulation(self, y) for y in x] for x in overrides_formatted
        ]
        self.holdout: Optional[HoldoutExperiment] = (
            HoldoutExperiment(self, holdout_name, holdout_bucket)
            if holdout_name is not None and holdout_bucket is not None
            else None
        )
        self.aa_mode: bool = aa_mode == 1
        self.trigger_debugging: bool = trigger_debugging == 1

    def __repr__(self) -> str:
        return f'<GuildExperiment hash={self.hash}{f" name={self._name!r}" if self._name else ""}>'

    def __hash__(self) -> int:
        return self.hash

    def __eq__(self, other: object, /) -> bool:
        if isinstance(other, GuildExperiment):
            return self.hash == other.hash
        elif isinstance(other, UserExperiment):
            return False
        return NotImplemented

    @property
    def name(self) -> Optional[str]:
        """Optional[:class:`str`]: The unique name of the experiment.

        This data is not always available via the API, and must be set manually for using related functions.
        """
        return self._name

    @name.setter
    def name(self, value: Optional[str], /) -> None:
        if not value:
            self._name = None
        elif murmurhash32(value, signed=False) != self.hash:
            raise ValueError('The name provided does not match the experiment hash')
        else:
            self._name = value

    def result_for(self, guild: Snowflake, /) -> int:
        """Returns the calulated position of the guild within the experiment (0-9999).

        Parameters
        -----------
        guild: :class:`abc.Snowflake`
            The guild to compute the position for.

        Raises
        ------
        :exc:`ValueError`
            The experiment name is unset.

        Returns
        -------
        :class:`int`
            The position of the guild within the experiment.
        """
        if not self.name:
            raise ValueError('The experiment name must be set to compute the result')

        return murmurhash32(f'{self.name}:{guild.id}', signed=False) % 10000

    def bucket_for(self, guild: Guild, /, *, aa_mode: bool = False) -> int:
        """Returns the assigned experiment bucket for a guild.
        Defaults to None (-1) if the guild is not in the experiment.

        Parameters
        -----------
        guild: :class:`Guild`
            The guild to compute experiment eligibility for.
        aa_mode: :class:`bool`
            Whether to return the bucket for A/A mode.
            Otherwise, always returns None (-1) for populations in A/A mode.
            Does nothing if the experiment is not in A/A mode.

        Raises
        ------
        :exc:`ValueError`
            The experiment name is unset.

        Returns
        -------
        :class:`int`
            The experiment bucket.
        """
        # Overrides take precedence
        for override in self.overrides:
            if guild.id in override.ids:
                return override.bucket

        hash_result = self.result_for(guild)
        for overrides in self.overrides_formatted:
            for override in overrides:
                if override.is_eligible(guild):
                    return override.bucket_for(guild, hash_result)

        # a/a mode is always -1 without an override
        if self.aa_mode and not aa_mode:
            return -1

        # Holdout must not be fulfilled
        if self.holdout and self.holdout.is_blocked(guild):
            return -1

        for population in self.populations:
            if population.is_eligible(guild):
                return population.bucket_for(guild, hash_result)

        return -1

    def guilds_for(self, bucket: int, /) -> List[Guild]:
        """Returns a list of guilds assigned to a specific bucket.

        Parameters
        -----------
        bucket: :class:`int`
            The bucket to get guilds for.

        Raises
        ------
        :exc:`ValueError`
            The experiment name is unset.

        Returns
        -------
        List[:class:`Guild`]
            The guilds assigned to the bucket.
        """
        return [x for x in self._state.guilds if self.bucket_for(x) == bucket]


class UserExperiment:
    """Represents a user's experiment assignment.

    .. container:: operations

        .. describe:: x == y

            Checks if two experiments are equal.

        .. describe:: x != y

            Checks if two experiments are not equal.

        .. describe:: hash(x)

            Returns the experiment's hash.

    .. note::

        In contrast to the wide range of data provided for guild experiments,
        user experiments do not reveal detailed rollout information, providing only the assigned bucket.

    .. versionadded:: 2.1

    Attributes
    ----------
    hash: :class:`int`
        The 32-bit unsigned Murmur3 hash of the experiment's name.
    revision: :class:`int`
        The current revision of the experiment rollout.
    assignment: :class:`int`
        The assigned bucket for the user.
    override: :class:`bool`
        Whether the user has an explicit bucket override.
    population: :class:`int`
        The internal population group for the user, or None (-1) if manually overridden.
    holdout: Optional[:class:`UserExperiment`]
        An experiment that blocks the rollout of this experiment.
        Only present if the user has an assignment in the holdout.
    aa_mode: :class:`bool`
        Whether the experiment is in A/A mode.
    trigger_debugging: :class:`bool`
        Whether experiment analytics trigger debugging is enabled.
    """

    __slots__ = (
        '_state',
        '_name',
        'hash',
        'revision',
        'assignment',
        'override',
        'population',
        '_result',
        'aa_mode',
        'trigger_debugging',
        'holdout',
    )

    def __init__(self, *, state: ConnectionState, data: AssignmentPayload):
        (
            hash,
            revision,
            bucket,
            override,
            population,
            hash_result,
            aa_mode,
            trigger_debugging,
            holdout_name,
            holdout_revision,
            holdout_bucket,
            *_,
        ) = data

        self._state = state
        self._name: Optional[str] = None
        self.hash: int = hash
        self.revision: int = revision
        self.assignment: int = bucket
        self.override: bool = override == 0  # this is fucking weird
        self.population: int = population
        self._result: int = hash_result
        self.aa_mode: bool = aa_mode == 1
        self.trigger_debugging: bool = trigger_debugging == 1

        self.holdout: Optional[UserExperiment] = None
        if holdout_name is not None:
            holdout_hash = murmurhash32(holdout_name, signed=False)
            self.holdout = state.experiments.get(holdout_hash) or UserExperiment.from_holdout(
                state, holdout_hash, holdout_name, holdout_revision, holdout_bucket
            )

    @classmethod
    def from_holdout(
        cls, state: ConnectionState, hash: int, name: str, revision: Optional[int], bucket: Optional[int]
    ) -> UserExperiment:
        self = cls.__new__(cls)
        self._state = state
        self._name = name
        self.hash = hash
        self.revision = revision or 0
        self.assignment = bucket or -1

        # Most of these fields are defaults
        self.override = False
        self.population = 0
        self._result = -1
        self.aa_mode = False
        self.trigger_debugging = False
        self.holdout = None
        return self

    def __repr__(self) -> str:
        return f'<UserExperiment hash={self.hash}{f" name={self._name!r}" if self._name else ""} bucket={self.bucket}>'

    def __hash__(self) -> int:
        return self.hash

    def __eq__(self, other: object, /) -> bool:
        if isinstance(other, UserExperiment):
            return self.hash == other.hash
        elif isinstance(other, GuildExperiment):
            return False
        return NotImplemented

    @property
    def name(self) -> Optional[str]:
        """Optional[:class:`str`]: The unique name of the experiment.

        This data is not always available via the API, and must be set manually for using related functions.
        """
        return self._name

    @name.setter
    def name(self, value: Optional[str], /) -> None:
        if not value:
            self._name = None
        elif murmurhash32(value, signed=False) != self.hash:
            raise ValueError('The name provided does not match the experiment hash')
        else:
            self._name = value

    @property
    def bucket(self) -> int:
        """:class:`int`: The assigned bucket for the user."""
        if self.aa_mode:
            return -1
        return self.assignment

    @property
    def result(self) -> int:
        """:class:`int`: The calulated position of the user within the experiment (0-9999).

        Raises
        ------
        :exc:`ValueError`
            The experiment name is unset without a precomputed result.
        """
        if self._result >= 0:
            return self._result
        elif not self.name:
            raise ValueError('The experiment name must be set to compute the result')
        else:
            result = murmurhash32(f'{self.name}:{self._state.self_id}', signed=False) % 10000
            self._result = result
            return result



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import List, Literal, TypedDict, Union
from typing_extensions import NotRequired

from .application import IntegrationApplication
from .command import ApplicationCommand
from .components import ModalActionRow
from .member import Member
from .message import PartialAttachment
from .snowflake import Snowflake
from .user import User

InteractionType = Literal[1, 2, 3, 4, 5]


class _BaseApplicationCommandInteractionDataOption(TypedDict):
    name: str


class _CommandGroupApplicationCommandInteractionDataOption(_BaseApplicationCommandInteractionDataOption):
    type: Literal[1, 2]
    options: List[ApplicationCommandInteractionDataOption]


class _BaseValueApplicationCommandInteractionDataOption(_BaseApplicationCommandInteractionDataOption, total=False):
    focused: bool


class _StringValueApplicationCommandInteractionDataOption(_BaseValueApplicationCommandInteractionDataOption):
    type: Literal[3]
    value: str


class _IntegerValueApplicationCommandInteractionDataOption(_BaseValueApplicationCommandInteractionDataOption):
    type: Literal[4]
    value: int


class _BooleanValueApplicationCommandInteractionDataOption(_BaseValueApplicationCommandInteractionDataOption):
    type: Literal[5]
    value: bool


class _SnowflakeValueApplicationCommandInteractionDataOption(_BaseValueApplicationCommandInteractionDataOption):
    type: Literal[6, 7, 8, 9, 11]
    value: Snowflake


class _NumberValueApplicationCommandInteractionDataOption(_BaseValueApplicationCommandInteractionDataOption):
    type: Literal[10]
    value: float


_ValueApplicationCommandInteractionDataOption = Union[
    _StringValueApplicationCommandInteractionDataOption,
    _IntegerValueApplicationCommandInteractionDataOption,
    _BooleanValueApplicationCommandInteractionDataOption,
    _SnowflakeValueApplicationCommandInteractionDataOption,
    _NumberValueApplicationCommandInteractionDataOption,
]


ApplicationCommandInteractionDataOption = Union[
    _CommandGroupApplicationCommandInteractionDataOption,
    _ValueApplicationCommandInteractionDataOption,
]


class _BaseApplicationCommandInteractionData(TypedDict):
    id: Snowflake
    name: str
    version: Snowflake
    guild_id: NotRequired[Snowflake]
    application_command: ApplicationCommand
    attachments: List[PartialAttachment]
    options: List[ApplicationCommandInteractionDataOption]


class ChatInputCommandInteractionData(_BaseApplicationCommandInteractionData, total=False):
    type: Literal[1]


class _BaseNonChatInputApplicationCommandInteractionData(_BaseApplicationCommandInteractionData):
    target_id: Snowflake


class UserCommandInteractionData(_BaseNonChatInputApplicationCommandInteractionData):
    type: Literal[2]


class MessageCommandInteractionData(_BaseNonChatInputApplicationCommandInteractionData):
    type: Literal[3]


ApplicationCommandInteractionData = Union[
    ChatInputCommandInteractionData,
    UserCommandInteractionData,
    MessageCommandInteractionData,
]


class _BaseMessageComponentInteractionData(TypedDict):
    custom_id: str


class ButtonInteractionData(_BaseMessageComponentInteractionData):
    component_type: Literal[2]


class SelectInteractionData(_BaseMessageComponentInteractionData):
    component_type: Literal[3]
    values: List[str]


MessageComponentInteractionData = Union[ButtonInteractionData, SelectInteractionData]


class TextInputInteractionData(TypedDict):
    type: Literal[4]
    custom_id: str
    value: str


ModalSubmitComponentInteractionData = TextInputInteractionData


class MessageActionRowData(TypedDict):
    type: Literal[1]
    components: List[MessageComponentInteractionData]


class ModalSubmitActionRowData(TypedDict):
    type: Literal[1]
    components: List[ModalSubmitComponentInteractionData]


class ModalSubmitInteractionData(TypedDict):
    id: Snowflake
    custom_id: str
    components: List[ModalSubmitComponentInteractionData]


ActionRowInteractionData = Union[MessageActionRowData, ModalSubmitActionRowData]
ComponentInteractionData = Union[MessageComponentInteractionData, ModalSubmitComponentInteractionData]
InteractionData = Union[
    ApplicationCommandInteractionData,
    MessageComponentInteractionData,
    ModalSubmitInteractionData,
]


class MessageInteraction(TypedDict):
    id: Snowflake
    type: InteractionType
    name: str
    user: User
    member: NotRequired[Member]


class Modal(TypedDict):
    id: int
    nonce: NotRequired[Snowflake]
    channel_id: Snowflake
    title: str
    custom_id: str
    application: IntegrationApplication
    components: List[ModalActionRow]



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import Dict, List, Literal, Optional, TypedDict, Union
from typing_extensions import NotRequired

from .command import ApplicationCommand
from .guild import PartialGuild
from .snowflake import Snowflake
from .team import Team
from .user import APIUser, PartialUser


class Token(TypedDict):
    token: str


class OptionalToken(TypedDict):
    # Missing if a bot already exists 😭
    token: Optional[str]


class Secret(TypedDict):
    secret: str


class BaseApplication(TypedDict):
    id: Snowflake
    name: str
    description: str
    icon: Optional[str]
    cover_image: NotRequired[str]
    splash: NotRequired[str]
    type: Optional[int]
    primary_sku_id: NotRequired[Snowflake]
    summary: NotRequired[Literal['']]
    deeplink_uri: NotRequired[str]
    third_party_skus: NotRequired[List[ThirdPartySKU]]
    bot: NotRequired[PartialUser]
    is_verified: bool
    is_discoverable: bool
    is_monetized: bool
    storefront_available: bool


class DetectableApplication(TypedDict):
    id: Snowflake
    name: str
    hook: bool
    overlay: NotRequired[bool]
    overlay_methods: NotRequired[int]
    overlay_warn: NotRequired[bool]
    overlay_compatibility_hook: NotRequired[bool]
    aliases: NotRequired[List[str]]
    executables: NotRequired[List[ApplicationExecutable]]


class IntegrationApplication(BaseApplication):
    role_connections_verification_url: NotRequired[Optional[str]]


class PartialApplication(BaseApplication, DetectableApplication):
    owner: NotRequired[APIUser]  # Not actually ever present in partial app
    team: NotRequired[Team]
    verify_key: str
    flags: int
    rpc_origins: NotRequired[List[str]]
    terms_of_service_url: NotRequired[str]
    privacy_policy_url: NotRequired[str]
    max_participants: NotRequired[Optional[int]]
    bot_public: NotRequired[bool]
    bot_require_code_grant: NotRequired[bool]
    integration_public: NotRequired[bool]
    integration_require_code_grant: NotRequired[bool]
    guild_id: NotRequired[Snowflake]
    primary_sku_id: NotRequired[Snowflake]
    slug: NotRequired[str]
    developers: NotRequired[List[Company]]
    publishers: NotRequired[List[Company]]
    eula_id: NotRequired[Snowflake]
    embedded_activity_config: NotRequired[EmbeddedActivityConfig]
    guild: NotRequired[PartialGuild]
    install_params: NotRequired[ApplicationInstallParams]
    store_listing_sku_id: NotRequired[Snowflake]


class ApplicationDiscoverability(TypedDict):
    discoverability_state: int
    discovery_eligibility_flags: int
    bad_commands: List[ApplicationCommand]


InteractionsVersion = Literal[1, 2]


class Application(PartialApplication, IntegrationApplication):
    bot_disabled: NotRequired[bool]
    bot_quarantined: NotRequired[bool]
    redirect_uris: List[str]
    interactions_endpoint_url: Optional[str]
    interactions_version: InteractionsVersion
    interactions_event_types: List[str]
    verification_state: int
    store_application_state: int
    rpc_application_state: int
    creator_monetization_state: int
    discoverability_state: int
    discovery_eligibility_flags: int
    monetization_state: int
    monetization_eligibility_flags: int
    approximate_guild_count: NotRequired[int]


class WhitelistedUser(TypedDict):
    user: PartialUser
    state: Literal[1, 2]


class Asset(TypedDict):
    id: Snowflake
    name: str
    type: int


class StoreAsset(TypedDict):
    id: Snowflake
    size: int
    width: int
    height: int
    mime_type: str


class Company(TypedDict):
    id: Snowflake
    name: str


class EULA(TypedDict):
    id: Snowflake
    name: str
    content: str


class ApplicationExecutable(TypedDict):
    name: str
    os: Literal['win32', 'linux', 'darwin']
    is_launcher: bool


class ThirdPartySKU(TypedDict):
    distributor: Literal['discord', 'steam', 'twitch', 'uplay', 'battlenet', 'origin', 'gog', 'epic', 'google_play']
    id: Optional[str]
    sku_id: Optional[str]


class BaseAchievement(TypedDict):
    id: Snowflake
    name: Union[str, Dict[str, Union[str, Dict[str, str]]]]
    name_localizations: NotRequired[Dict[str, str]]
    description: Union[str, Dict[str, Union[str, Dict[str, str]]]]
    description_localizations: NotRequired[Dict[str, str]]
    icon_hash: str
    secure: bool
    secret: bool


class Achievement(BaseAchievement):
    application_id: Snowflake


class Ticket(TypedDict):
    ticket: str


class Branch(TypedDict):
    id: Snowflake
    live_build_id: NotRequired[Optional[Snowflake]]
    created_at: NotRequired[str]
    name: NotRequired[str]


class BranchSize(TypedDict):
    size_kb: str  # Stringified float


class DownloadSignature(TypedDict):
    endpoint: str
    expires: int
    signature: str


class Build(TypedDict):
    application_id: NotRequired[Snowflake]
    created_at: NotRequired[str]
    id: Snowflake
    manifests: List[Manifest]
    status: Literal['CORRUPTED', 'INVALID', 'READY', 'VALIDATING', 'UPLOADED', 'UPLOADING', 'CREATED']
    source_build_id: NotRequired[Optional[Snowflake]]
    version: NotRequired[Optional[str]]


class CreatedBuild(TypedDict):
    build: Build
    manifest_uploads: List[Manifest]


class BuildFile(TypedDict):
    id: Snowflake
    md5_hash: NotRequired[str]


class CreatedBuildFile(TypedDict):
    id: str
    url: str


class ManifestLabel(TypedDict):
    application_id: Snowflake
    id: Snowflake
    name: NotRequired[str]


class Manifest(TypedDict):
    id: Snowflake
    label: ManifestLabel
    redistributable_label_ids: NotRequired[List[Snowflake]]
    url: Optional[str]


class _BaseActivityStatistics(TypedDict):
    total_duration: int
    last_played_at: str


class UserActivityStatistics(_BaseActivityStatistics):
    application_id: Snowflake
    total_discord_sku_duration: NotRequired[int]
    first_played_at: Optional[str]


class ApplicationActivityStatistics(_BaseActivityStatistics):
    user_id: Snowflake


class GlobalActivityStatistics(TypedDict):
    application_id: Snowflake
    application: NotRequired[PartialApplication]
    user_id: Snowflake
    user: NotRequired[PartialUser]
    duration: int
    updated_at: str


EmbeddedActivityPlatform = Literal['web', 'android', 'ios']
EmbeddedActivityPlatformReleasePhase = Literal[
    'in_development', 'activities_team', 'employee_release', 'soft_launch', 'global_launch'
]


class EmbeddedActivityPlatformConfig(TypedDict):
    label_type: Literal[0, 1, 2]
    label_until: Optional[str]
    release_phase: EmbeddedActivityPlatformReleasePhase


class EmbeddedActivityConfig(TypedDict):
    application_id: NotRequired[Snowflake]
    activity_preview_video_asset_id: NotRequired[Optional[Snowflake]]
    client_platform_config: Dict[EmbeddedActivityPlatform, EmbeddedActivityPlatformConfig]
    default_orientation_lock_state: Literal[1, 2, 3]
    tablet_default_orientation_lock_state: Literal[1, 2, 3]
    free_period_ends_at: NotRequired[Optional[str]]
    free_period_starts_at: NotRequired[Optional[str]]
    premium_tier_requirement: NotRequired[Optional[Literal[1, 2, 3]]]
    requires_age_gate: bool
    shelf_rank: int
    supported_platforms: List[EmbeddedActivityPlatform]


class ApplicationInstallParams(TypedDict):
    scopes: List[str]
    permissions: int


class ActiveDeveloperWebhook(TypedDict):
    channel_id: Snowflake
    webhook_id: Snowflake


class ActiveDeveloperResponse(TypedDict):
    follower: ActiveDeveloperWebhook


class RoleConnectionMetadata(TypedDict):
    type: Literal[1, 2, 3, 4, 5, 6, 7, 8]
    key: str
    name: str
    description: str
    name_localizations: NotRequired[Dict[str, str]]
    description_localizations: NotRequired[Dict[str, str]]


class PartialRoleConnection(TypedDict):
    platform_name: Optional[str]
    platform_username: Optional[str]
    metadata: Dict[str, str]


class RoleConnection(PartialRoleConnection):
    application: BaseApplication
    application_metadata: List[RoleConnectionMetadata]


class UnverifiedApplication(TypedDict):
    name: str
    hash: str
    missing_data: List[str]


class ExternalAsset(TypedDict):
    url: str
    external_asset_path: str



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import Any, Dict, List, Literal, Optional, TypedDict
from typing_extensions import NotRequired

from .integration import ConnectionIntegration
from .snowflake import Snowflake

NameplatePallete = Literal['crimson', 'berry', 'sky', 'teal', 'forest', 'bubble_gum', 'violet', 'cobalt', 'clover']


class _UserSKU(TypedDict):
    asset: str
    sku_id: Snowflake


class AvatarDecorationData(_UserSKU):
    expires_at: Optional[int]


class Collectible(_UserSKU):
    label: str
    expires_at: Optional[str]


class NameplateCollectible(Collectible):
    palette: str


class UserCollectibles(TypedDict):
    nameplate: NameplateCollectible


class PartialUser(TypedDict):
    id: Snowflake
    username: str
    discriminator: str
    avatar: Optional[str]
    avatar_decoration_data: NotRequired[Optional[AvatarDecorationData]]
    public_flags: NotRequired[int]
    bot: NotRequired[bool]
    system: NotRequired[bool]
    global_name: Optional[str]
    primary_guild: NotRequired[Optional[PrimaryGuild]]
    display_name_styles: NotRequired[Optional[DisplayNameStyle]]
    collectibles: NotRequired[Optional[UserCollectibles]]


ConnectionType = Literal[
    'battlenet',
    'contacts',
    'crunchyroll',
    'ebay',
    'epicgames',
    'facebook',
    'github',
    'instagram',
    'leagueoflegends',
    'paypal',
    'playstation',
    'reddit',
    'riotgames',
    'samsung',
    'spotify',
    'skype',
    'steam',
    'tiktok',
    'twitch',
    'twitter',
    'youtube',
    'xbox',
]
ConnectionVisibilty = Literal[0, 1]
PremiumType = Literal[0, 1, 2, 3]
DisplayNameFont = Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
DisplayNameEffect = Literal[1, 2, 3, 4, 5, 6]


class APIUser(PartialUser):
    banner: Optional[str]
    accent_color: Optional[int]


class User(APIUser, total=False):
    mfa_enabled: bool
    locale: str
    verified: bool
    email: Optional[str]
    flags: int
    purchased_flags: int
    premium_usage_flags: int
    premium_type: PremiumType
    bio: str
    analytics_token: str
    phone: Optional[str]
    token: str
    nsfw_allowed: Optional[bool]
    desktop: bool
    mobile: bool


class DisplayNameStyle(TypedDict):
    font_id: DisplayNameFont
    effect_id: DisplayNameEffect
    colors: List[int]  # 1-2


class UserWithToken(User):
    token: str


class PrimaryGuild(TypedDict):
    identity_guild_id: Optional[int]
    identity_enabled: Optional[bool]
    tag: Optional[str]
    badge: Optional[str]


class PomeloAttempt(TypedDict):
    taken: bool


class PomeloSuggestion(TypedDict):
    username: str


class PartialConnection(TypedDict):
    id: str
    type: ConnectionType
    name: str
    verified: bool
    metadata: NotRequired[Dict[str, Any]]


class Connection(PartialConnection):
    revoked: bool
    visibility: Literal[0, 1]
    metadata_visibility: Literal[0, 1]
    show_activity: bool
    friend_sync: bool
    two_way_link: bool
    integrations: NotRequired[List[ConnectionIntegration]]
    access_token: NotRequired[str]


class ConnectionAccessToken(TypedDict):
    access_token: str


class ConnectionAuthorization(TypedDict):
    url: str


RelationshipType = Literal[-1, 0, 1, 2, 3, 4, 5, 6]


class Relationship(TypedDict):
    id: Snowflake
    type: RelationshipType
    user: PartialUser
    nickname: Optional[str]
    since: NotRequired[str]


ProtoSettingsType = Literal[1, 2, 3]


class ProtoSettings(TypedDict):
    settings: str
    out_of_date: NotRequired[bool]


class _ConsentSettings(TypedDict):
    consented: bool


class PartialConsentSettings(TypedDict):
    personalization: _ConsentSettings


class ConsentSettings(PartialConsentSettings):
    usage_statistics: _ConsentSettings


class _EmailSettingsCategories(TypedDict):
    communication: bool
    social: bool
    recommendations_and_events: bool
    tips: bool
    updates_and_announcements: bool
    family_center_digest: bool


class EmailSettings(TypedDict):
    initialized: bool
    categories: _EmailSettingsCategories


MessageNotificationLevel = Literal[0, 1, 2, 3]
HighlightLevel = Literal[0, 1, 2]


class MuteConfig(TypedDict):
    end_time: Optional[str]
    selected_time_window: Optional[int]


class ChannelOverride(TypedDict):
    channel_id: Snowflake
    collapsed: bool
    message_notifications: MessageNotificationLevel
    muted: bool
    mute_config: Optional[MuteConfig]


class UserGuildSettings(TypedDict):
    guild_id: Optional[Snowflake]
    channel_overrides: List[ChannelOverride]
    flags: int
    message_notifications: MessageNotificationLevel
    notify_highlights: HighlightLevel
    hide_muted_channels: bool
    mobile_push: bool
    muted: bool
    mute_config: Optional[MuteConfig]
    mute_scheduled_events: bool
    suppress_everyone: bool
    suppress_roles: bool
    version: int


class UserAffinity(TypedDict):
    user_id: Snowflake
    affinity: float


class UserAffinities(TypedDict):
    user_affinities: List[UserAffinity]
    inverse_user_affinities: List[UserAffinity]


class GuildAffinity(TypedDict):
    guild_id: Snowflake
    affinity: float


class GuildAffinities(TypedDict):
    guild_affinities: List[GuildAffinity]


class ChannelAffinity(TypedDict):
    channel_id: Snowflake
    affinity: float


class ChannelAffinities(TypedDict):
    channel_affinities: List[ChannelAffinity]


class Note(TypedDict):
    note: str
    user_id: Snowflake
    note_user_id: Snowflake


class FriendSuggestionReason(TypedDict):
    name: str
    platform_type: ConnectionType
    type: Literal[1]


class FriendSuggestion(TypedDict):
    suggested_user: PartialUser
    reasons: List[FriendSuggestionReason]
    from_suggested_user_contacts: NotRequired[bool]


class Report(TypedDict):
    report_id: Snowflake


class UserAvatar(TypedDict):
    id: Snowflake
    storage_hash: str
    description: Optional[str]


class UserAvatars(TypedDict):
    avatars: List[UserAvatar]



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import asyncio
from collections import deque
import logging
import struct
import time
import threading
import traceback

from typing import Any, Callable, Coroutine, Dict, List, TYPE_CHECKING, NamedTuple, Optional, Sequence, TypeVar, Tuple

from curl_cffi import CurlError, WebSocketError
from curl_cffi.requests import AsyncWebSocket
from curl_cffi.const import CurlWsFlag
import yarl

from . import utils
from .enums import SpeakingState, Status
from .errors import ClientException, ConnectionClosed
from .flags import Capabilities

try:
    import davey  # type: ignore
except ImportError:
    pass

_log = logging.getLogger(__name__)

__all__ = (
    'DiscordWebSocket',
    'KeepAliveHandler',
    'VoiceKeepAliveHandler',
    'DiscordVoiceWebSocket',
    'ReconnectWebSocket',
    'ConnectionClosed',
)

if TYPE_CHECKING:
    from typing_extensions import Self

    from .client import Client
    from .state import ConnectionState
    from .types.activity import Activity as ActivityPayload
    from .types.snowflake import Snowflake
    from .types.gateway import BulkGuildSubscribePayload
    from .voice_state import VoiceConnectionState


class ReconnectWebSocket(Exception):
    """Signals to safely reconnect the websocket."""

    def __init__(self, *, resume: bool = True):
        self.resume = resume
        self.op: str = 'RESUME' if resume else 'IDENTIFY'


class WebSocketClosure(Exception):
    """An exception to make up for the fact that curl doesn't signal closure.

    Attributes
    -----------
    code: :class:`int`
        The close code of the websocket.
    reason: :class:`str`
        The reason provided for the closure.
    """

    __slots__ = ('code', 'reason')

    def __init__(self, socket: AsyncWebSocket):
        self.code: int = socket.close_code or -1
        self.reason: str = socket.close_reason or ''
        super().__init__(f'Websocket closed with {self.code} (reason: {self.reason!r})')


class EventListener(NamedTuple):
    predicate: Callable[[Dict[str, Any]], bool]
    event: str
    result: Optional[Callable[[Dict[str, Any]], Any]]
    future: asyncio.Future[Any]


class GatewayRatelimiter:
    def __init__(self, count: int = 110, per: float = 60.0) -> None:
        # The default is 110 to give room for at least 10 heartbeats per minute
        self.max: int = count
        self.remaining: int = count
        self.window: float = 0.0
        self.per: float = per
        self.lock: asyncio.Lock = asyncio.Lock()

    def is_ratelimited(self) -> bool:
        current = time.time()
        if current > self.window + self.per:
            return False
        return self.remaining == 0

    def get_delay(self) -> float:
        current = time.time()

        if current > self.window + self.per:
            self.remaining = self.max

        if self.remaining == self.max:
            self.window = current

        if self.remaining == 0:
            return self.per - (current - self.window)

        self.remaining -= 1
        return 0.0

    async def block(self) -> None:
        async with self.lock:
            delta = self.get_delay()
            if delta:
                _log.warning('Gateway is ratelimited, waiting %.2f seconds.', delta)
                await asyncio.sleep(delta)


class KeepAliveHandler:  # Inspired by enhanced-discord.py/Gnome
    def __init__(self, *, ws: DiscordWebSocket, interval: Optional[float] = None):
        self.ws: DiscordWebSocket = ws
        self.interval: Optional[float] = interval
        self.heartbeat_timeout: float = self.ws._max_heartbeat_timeout

        self.msg: str = 'Keeping websocket alive.'
        self.block_msg: str = 'Heartbeat blocked for more than %s seconds.'
        self.behind_msg: str = "Can't keep up, websocket is %.1fs behind."
        self.not_responding_msg: str = 'Gateway has stopped responding. Closing and restarting.'
        self.no_stop_msg: str = 'An error occurred while stopping the Gateway. Ignoring.'

        self._stop: asyncio.Event = asyncio.Event()
        self._last_send: float = time.perf_counter()
        self._last_recv: float = time.perf_counter()
        self._last_ack: float = time.perf_counter()
        self.latency: float = float('inf')

    async def run(self) -> None:
        while True:
            try:
                await asyncio.wait_for(self._stop.wait(), timeout=self.interval)
            except asyncio.TimeoutError:
                pass
            else:
                return

            if self._last_recv + self.heartbeat_timeout < time.perf_counter():
                _log.warning(self.not_responding_msg)

                try:
                    await self.ws.close(4000)
                except Exception:
                    _log.exception(self.no_stop_msg)
                finally:
                    self.stop()
                return

            data = self.get_payload()
            _log.debug(self.msg)
            try:
                total = 0
                while True:
                    try:
                        await asyncio.wait_for(self.ws.send_heartbeat(data), timeout=10)
                        break
                    except asyncio.TimeoutError:
                        total += 10

                        stack = ''.join(traceback.format_stack())
                        msg = f'{self.block_msg}\nLoop traceback (most recent call last):\n{stack}'
                        _log.warning(msg, total)

            except Exception:
                self.stop()
            else:
                self._last_send = time.perf_counter()

    def get_payload(self) -> Dict[str, Any]:
        return {
            'op': self.ws.HEARTBEAT,
            'd': self.ws.sequence,
        }

    def start(self) -> None:
        self.ws.loop.create_task(self.run())

    def stop(self) -> None:
        self._stop.set()

    def tick(self) -> None:
        self._last_recv = time.perf_counter()

    def ack(self) -> None:
        ack_time = time.perf_counter()
        self._last_ack = ack_time
        self.latency = ack_time - self._last_send
        if self.latency > 10:
            _log.warning(self.behind_msg, self.latency)


class VoiceKeepAliveHandler(KeepAliveHandler):
    if TYPE_CHECKING:
        ws: DiscordVoiceWebSocket

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.recent_ack_latencies: deque[float] = deque(maxlen=20)
        self.msg: str = 'Keeping voice socket alive.'
        self.block_msg: str = 'Voice heartbeat blocked for more than %s seconds'
        self.behind_msg: str = 'High voice socket latency, heartbeat is %.1fs behind'
        self.not_responding_msg: str = 'Voice socket has stopped responding. Closing and restarting.'
        self.no_stop_msg: str = 'An error occurred while stopping the voice socket. Ignoring.'

    def get_payload(self) -> Dict[str, Any]:
        return {
            'op': self.ws.HEARTBEAT,
            'd': {
                't': int(time.time() * 1000),
                'seq_ack': self.ws.seq_ack,
            },
        }

    def ack(self) -> None:
        ack_time = time.perf_counter()
        self._last_ack = ack_time
        self._last_recv = ack_time
        self.latency: float = ack_time - self._last_send
        self.recent_ack_latencies.append(self.latency)
        if self.latency > 10:
            _log.warning(self.behind_msg, self.latency)


DWS = TypeVar('DWS', bound='DiscordWebSocket')


class DiscordWebSocket:
    """Implements a WebSocket for Discord's Gateway v9.

    Attributes
    -----------
    gateway
        The Gateway we are currently connected to.
    token
        The authentication token for Discord.
    """

    if TYPE_CHECKING:
        token: Optional[str]
        _connection: ConnectionState
        _discord_parsers: Dict[str, Callable[..., Any]]
        call_hooks: Callable[..., Any]
        _initial_identify: bool
        shard_id: Optional[int]
        shard_count: Optional[int]
        gateway: yarl.URL
        _max_heartbeat_timeout: float
        _headers: utils.Headers
        _transport_compression: bool

    # fmt: off
    DEFAULT_GATEWAY       = yarl.URL('wss://gateway.discord.gg/')
    DISPATCH              = 0
    HEARTBEAT             = 1
    IDENTIFY              = 2
    PRESENCE              = 3
    VOICE_STATE           = 4
    VOICE_PING            = 5
    RESUME                = 6
    RECONNECT             = 7
    REQUEST_MEMBERS       = 8
    INVALIDATE_SESSION    = 9
    HELLO                 = 10
    HEARTBEAT_ACK         = 11
    # GUILD_SYNC          = 12
    CALL_CONNECT          = 13
    GUILD_SUBSCRIBE       = 14  # Deprecated
    # REQUEST_COMMANDS    = 24
    SEARCH_RECENT_MEMBERS = 35
    BULK_GUILD_SUBSCRIBE  = 37
    # fmt: on

    def __init__(self, socket: AsyncWebSocket, *, loop: asyncio.AbstractEventLoop) -> None:
        self.socket: AsyncWebSocket = socket
        self.loop: asyncio.AbstractEventLoop = loop

        # An empty dispatcher to prevent crashes
        self._dispatch: Callable[..., Any] = lambda *args: None
        # Generic event listeners
        self._dispatch_listeners: List[EventListener] = []
        # The keep alive
        self._keep_alive: Optional[KeepAliveHandler] = None
        self.thread_id: int = threading.get_ident()

        # WS related stuff
        self.session_id: Optional[str] = None
        self.sequence: Optional[int] = None
        self._decompressor: utils._DecompressionContext = utils._ActiveDecompressionContext()
        self._close_code: Optional[int] = None
        self._rate_limiter: GatewayRatelimiter = GatewayRatelimiter()

        self._hello_trace: List[str] = []
        self._session_trace: List[str] = []
        self._resume_trace: List[str] = []
        self._initial_identify: bool = False

        # Presence state tracking
        self.status: str = Status.unknown.value
        self.activities: List[ActivityPayload] = []
        self.afk: bool = False
        self.idle_since: int = 0

    @property
    def open(self) -> bool:
        return not self.socket.closed

    @property
    def capabilities(self) -> Capabilities:
        return Capabilities.default()

    def is_ratelimited(self) -> bool:
        return self._rate_limiter.is_ratelimited()

    def debug_log_receive(self, data: Dict[str, Any], /) -> None:
        self._dispatch('socket_raw_receive', data)

    def log_receive(self, _: Dict[str, Any], /) -> None:
        pass

    @classmethod
    async def from_client(
        cls,
        client: Client,
        *,
        initial: bool = False,
        gateway: Optional[yarl.URL] = None,
        session: Optional[str] = None,
        sequence: Optional[int] = None,
        resume: bool = False,
        encoding: str = 'json',
        compress: bool = True,
    ) -> Self:
        """Creates a main websocket for Discord from a :class:`Client`.

        This is for internal use only.
        """
        # Circular import
        from .http import INTERNAL_API_VERSION

        gateway = gateway or cls.DEFAULT_GATEWAY

        if not compress:
            url = gateway.with_query(v=INTERNAL_API_VERSION, encoding=encoding)
        else:
            url = gateway.with_query(
                v=INTERNAL_API_VERSION, encoding=encoding, compress=utils._ActiveDecompressionContext.COMPRESSION_TYPE
            )

        socket = await client.http.ws_connect(str(url))
        ws = cls(socket, loop=client.loop)

        # Dynamically add attributes needed
        ws.token = client.http.token
        ws._connection = client._connection
        ws._discord_parsers = client._connection.parsers
        ws._dispatch = client.dispatch
        ws.gateway = gateway
        ws.call_hooks = client._connection.call_hooks
        ws._initial_identify = initial
        ws.session_id = session
        ws.sequence = sequence
        ws._max_heartbeat_timeout = client._connection.heartbeat_timeout
        ws._headers = client.http.headers
        ws._transport_compression = compress
        ws.afk = client._connection._afk
        ws.idle_since = client._connection._idle_since

        if client._enable_debug_events:
            ws.send = ws.debug_send
            ws.log_receive = ws.debug_log_receive

        client._connection._update_references(ws)
        _log.debug('Connected to %s.', gateway)

        if not resume:
            await ws.identify()
            return ws

        await ws.resume()
        return ws

    def wait_for(
        self,
        event: str,
        predicate: Callable[[Dict[str, Any]], bool],
        result: Optional[Callable[[Dict[str, Any]], Any]] = None,
    ) -> asyncio.Future[Any]:
        """Waits for a DISPATCH'd event that meets the predicate.

        Parameters
        -----------
        event: :class:`str`
            The event to wait for.
        predicate
            A function that takes a data parameter to check for event
            properties. The data parameter is the 'd' key in the JSON message.
        result
            A function that takes the same data parameter and executes to send
            the result to the future. If ``None``, returns the data.

        Returns
        --------
        asyncio.Future
            A future to wait for.
        """

        event = event.upper()
        future = self.loop.create_future()
        entry = EventListener(event=event, predicate=predicate, result=result, future=future)
        self._dispatch_listeners.append(entry)
        return future

    async def identify(self) -> None:
        """Sends the IDENTIFY packet."""

        # User presence is weird...
        # This payload is only sometimes respected; usually the gateway tells
        # us our presence through the READY packet's sessions key
        # However, when reidentifying, we should send our last known presence
        # initial_status and initial_activities could probably also be sent here
        # but that needs more testing...
        presence = {
            'status': 'unknown',
            'since': self.idle_since,
            'activities': [],
            'afk': self.afk,
        }
        existing = self._connection.current_session
        if existing is not None:
            presence['status'] = str(existing.status) if existing.status is not Status.offline else 'invisible'
            presence['activities'] = [a.to_dict() for a in existing.activities]
        # else:
        #     presence['status'] = self._connection._status or 'unknown'
        #     presence['activities'] = self._connection._activities

        properties = self._headers.gateway_properties

        payload = {
            'op': self.IDENTIFY,
            'd': {
                'token': self.token,
                'capabilities': self.capabilities.value,
                'properties': properties,
                'presence': presence,
                'compress': not self._transport_compression,  # We require at least one form of compression
                'client_state': {
                    'guild_versions': {},
                },
            },
        }

        await self.call_hooks('before_identify', initial=self._initial_identify)
        await self.send_as_json(payload)
        _log.debug('Gateway has sent the IDENTIFY payload.')
        self._initial_identify = True

    async def resume(self) -> None:
        """Sends the RESUME packet."""
        payload = {
            'op': self.RESUME,
            'd': {
                'seq': self.sequence,
                'session_id': self.session_id,
                'token': self.token,
            },
        }

        await self.send_as_json(payload)
        _log.debug('Gateway has sent the RESUME payload.')

    async def received_message(self, msg: Any, /) -> None:
        if type(msg) is bytes:
            msg = self._decompressor.decompress(msg)

            # Received a partial gateway message
            if msg is None:
                return

        self.log_receive(msg)
        msg = utils._from_json(msg)

        _log.debug('Gateway event: %s.', msg)
        event = msg.get('t')
        if event:
            self._dispatch('socket_event_type', event)

        op = msg.get('op')
        data = msg.get('d')
        seq = msg.get('s')
        if seq is not None:
            self.sequence = seq

        if self._keep_alive:
            self._keep_alive.tick()

        if op != self.DISPATCH:
            if op == self.RECONNECT:
                # RECONNECT can only be handled by the Client
                # so we terminate our connection and raise an
                # internal exception signalling to reconnect
                _log.debug('Received RECONNECT opcode.')
                await self.close()
                raise ReconnectWebSocket

            if op == self.HEARTBEAT_ACK:
                if self._keep_alive:
                    self._keep_alive.ack()
                return

            if op == self.HEARTBEAT:
                if self._keep_alive:
                    beat = self._keep_alive.get_payload()
                    await self.send_as_json(beat)
                return

            if op == self.HELLO:
                self._hello_trace = data.get('_trace', [])
                interval = data['heartbeat_interval'] / 1000.0
                self._keep_alive = KeepAliveHandler(ws=self, interval=interval)
                # Send a heartbeat immediately
                await self.send_as_json(self._keep_alive.get_payload())
                self._keep_alive.start()
                return

            if op == self.INVALIDATE_SESSION:
                if data is True:
                    await self.close()
                    raise ReconnectWebSocket

                self.sequence = None
                self.session_id = None
                self.gateway = self.DEFAULT_GATEWAY

                _log.info('Gateway session has been invalidated.')
                await self.close(code=1000)
                raise ReconnectWebSocket(resume=False)

            _log.warning('Unknown OP code %s.', op)
            return

        if event == 'READY':
            self._session_trace = data.get('_trace', [])
            self.sequence = msg['s']
            self.session_id = data['session_id']
            self.gateway = yarl.URL(data['resume_gateway_url'])

            _log.info('Connected to Gateway (session ID: %s).', self.session_id)
            await self.voice_state()  # Initial OP 4

        elif event == 'RESUMED':
            self._resume_trace = data.get('_trace', [])
            _log.info('Gateway has successfully RESUMED session %s.', self.session_id)

        try:
            func = self._discord_parsers[event]
        except KeyError:
            _log.debug('Unknown event %s.', event)
        else:
            try:
                func(data)
            except Exception as exc:
                if event in ('READY', 'READY_SUPPLEMENTAL'):
                    raise
                _log.warning(
                    'Parsing event %s encountered an exception. Please open an issue with this traceback:',
                    event,
                    exc_info=exc,
                )

        # Remove the dispatched listeners
        removed = []
        for index, entry in enumerate(self._dispatch_listeners):
            if entry.event != event:
                continue

            future = entry.future
            if future.cancelled():
                removed.append(index)
                continue

            try:
                valid = entry.predicate(data)
            except Exception as exc:
                future.set_exception(exc)
                removed.append(index)
            else:
                if valid:
                    ret = data if entry.result is None else entry.result(data)
                    future.set_result(ret)
                    removed.append(index)

        for index in reversed(removed):
            del self._dispatch_listeners[index]

    @property
    def latency(self) -> float:
        """:class:`float`: Measures latency between a HEARTBEAT and a HEARTBEAT_ACK in seconds."""
        heartbeat = self._keep_alive
        return float('inf') if heartbeat is None else heartbeat.latency

    def _can_handle_close(self, code: Optional[int] = None) -> bool:
        code = code or self._close_code
        # If the socket is closed remotely with 1000 and it's not our own explicit close
        # then it's an improper close that should be handled and reconnected
        is_improper_close = self._close_code is None and code == 1000
        return is_improper_close or code not in (1000, 4004, 4010, 4011, 4012, 4013, 4014)

    async def poll_event(self) -> None:
        """Polls for a DISPATCH event and handles the general Gateway loop.

        Raises
        ------
        ConnectionClosed
            The websocket connection was terminated for unhandled reasons.
        """
        try:
            msg, flags = await asyncio.wait_for(self.socket.recv(), timeout=self._max_heartbeat_timeout)
            if (flags & CurlWsFlag.TEXT) or (flags & CurlWsFlag.BINARY):
                await self.received_message(msg)
            elif flags & CurlWsFlag.CLOSE:
                socket = self.socket
                raise WebSocketClosure(socket)
        except (asyncio.TimeoutError, CurlError, WebSocketClosure) as e:
            _log.debug('Got Gateway poll exception: %s.', type(e), exc_info=True)
            # Ensure the keep alive handler is closed
            if self._keep_alive:
                self._keep_alive.stop()
                self._keep_alive = None

            if isinstance(e, asyncio.TimeoutError):
                _log.debug('Timed out receiving Gateway packet. Attempting a reconnect.')
                raise ReconnectWebSocket from None

            socket = self.socket
            code = self._close_code or socket.close_code
            reason = socket.close_reason
            if isinstance(e, CurlError):
                reason = str(e)

            if not socket.closed:
                await socket.close(code or 4000, (reason or 'Unknown error').encode('utf-8'))

            _log.info('Gateway received close code %s and reason %r.', code, reason)

            if self._can_handle_close(code or None):
                _log.debug('Websocket closed with %s, attempting a reconnect.', code)
                raise ReconnectWebSocket from None
            else:
                _log.debug('Websocket closed with %s, cannot reconnect.', code)
                raise ConnectionClosed(code, reason) from None
        except asyncio.CancelledError:
            if self._keep_alive:
                self._keep_alive.stop()
                self._keep_alive = None
            raise

    async def _sendstr(self, data: str, /, *, raise_on_closed: bool = False) -> None:
        try:
            await self.socket.send(data.encode('utf-8'))
        except WebSocketError:
            if self.socket.closed:
                # Not much we can do here
                _log.debug('Websocket is closed, cannot send data.')
                if raise_on_closed:
                    raise ClientException('WebSocket is closed')
            else:
                raise

    async def debug_send(self, data: str, /, *, raise_on_closed: bool = False) -> None:
        await self._rate_limiter.block()
        self._dispatch('socket_raw_send', data)
        await self._sendstr(data, raise_on_closed=raise_on_closed)

    async def send(self, data: str, /, *, raise_on_closed: bool = False) -> None:
        await self._rate_limiter.block()
        await self._sendstr(data, raise_on_closed=raise_on_closed)

    async def send_as_json(self, data: Any, /, *, raise_on_closed: bool = False) -> None:
        try:
            await self.send(utils._to_json(data), raise_on_closed=raise_on_closed)
        except RuntimeError as exc:
            if not self._can_handle_close(self._close_code):
                raise ConnectionClosed(self._close_code) from exc

    async def send_heartbeat(self, data: Any, /) -> None:
        # This bypasses the rate limit handling code since it has a higher priority
        try:
            await self._sendstr(utils._to_json(data))
        except RuntimeError as exc:
            if not self._can_handle_close(self._close_code):
                raise ConnectionClosed(self._close_code) from exc

    async def change_presence(
        self,
        *,
        activities: Optional[Sequence[ActivityPayload]] = None,
        status: str,
        since: int = 0,
        afk: bool = False,
    ) -> None:
        payload = {
            'op': self.PRESENCE,
            'd': {'activities': activities or [], 'afk': afk, 'since': since, 'status': str(status)},
        }

        _log.debug('Sending %s to change presence.', payload['d'])
        await self.send_as_json(payload)
        self.status = str(status)
        self.activities = list(activities or [])
        self.afk = afk
        self.idle_since = since

    async def guild_subscribe(
        self,
        guild_id: Snowflake,
        *,
        typing: Optional[bool] = None,
        threads: Optional[bool] = None,
        activities: Optional[bool] = None,
        members: Optional[List[Snowflake]] = None,
        channels: Optional[Dict[Snowflake, List[List[int]]]] = None,
        thread_member_lists: Optional[List[Snowflake]] = None,
    ):
        payload = {
            'op': self.GUILD_SUBSCRIBE,
            'd': {
                'guild_id': str(guild_id),
            },
        }

        data = payload['d']
        if typing is not None:
            data['typing'] = typing
        if threads is not None:
            data['threads'] = threads
        if activities is not None:
            data['activities'] = activities
        if members is not None:
            data['members'] = members
        if channels is not None:
            data['channels'] = channels
        if thread_member_lists is not None:
            data['thread_member_lists'] = thread_member_lists

        _log.debug('Subscribing to guild %s with payload %s', guild_id, payload['d'])
        await self.send_as_json(payload)

    async def bulk_guild_subscribe(self, subscriptions: BulkGuildSubscribePayload) -> None:
        payload = {
            'op': self.BULK_GUILD_SUBSCRIBE,
            'd': {
                'subscriptions': subscriptions,
            },
        }

        _log.debug('Subscribing to guilds with payload %s', payload['d'])
        await self.send_as_json(payload, raise_on_closed=True)

    async def request_chunks(
        self,
        guild_ids: List[Snowflake],
        query: Optional[str] = None,
        *,
        limit: Optional[int] = None,
        user_ids: Optional[List[Snowflake]] = None,
        presences: bool = True,
        nonce: Optional[str] = None,
    ) -> None:
        payload = {
            'op': self.REQUEST_MEMBERS,
            'd': {
                'guild_id': guild_ids,
                'query': query,
                'limit': limit,
                'presences': presences,
                'user_ids': user_ids,
            },
        }

        if nonce is not None:
            payload['d']['nonce'] = nonce

        await self.send_as_json(payload)

    async def voice_state(
        self,
        guild_id: Optional[int] = None,
        channel_id: Optional[int] = None,
        self_mute: bool = False,
        self_deaf: bool = False,
        self_video: bool = False,
    ) -> None:
        payload = {
            'op': self.VOICE_STATE,
            'd': {
                'guild_id': guild_id,
                'channel_id': channel_id,
                'self_mute': self_mute,
                'self_deaf': self_deaf,
                'self_video': self_video,
            },
        }

        if channel_id:
            payload['d'].update(self._connection._get_preferred_regions())

        _log.debug('Updating %s voice state to %s.', guild_id or 'client', payload)
        await self.send_as_json(payload)

    async def call_connect(self, channel_id: Snowflake):
        payload = {'op': self.CALL_CONNECT, 'd': {'channel_id': str(channel_id)}}

        _log.debug('Requesting call connect for channel %s.', channel_id)
        await self.send_as_json(payload)

    async def search_recent_members(
        self, guild_id: Snowflake, query: str = '', *, after: Optional[Snowflake] = None, nonce: Optional[str] = None
    ) -> None:
        payload = {
            'op': self.SEARCH_RECENT_MEMBERS,
            'd': {
                'guild_id': str(guild_id),
                'query': query,
                'continuation_token': str(after) if after else None,
            },
        }
        if nonce is not None:
            payload['d']['nonce'] = nonce

        await self.send_as_json(payload)

    async def close(self, code: int = 4000, reason: bytes = b'') -> None:
        _log.debug(f'Closing websocket with code {code}.')
        if self._keep_alive:
            self._keep_alive.stop()
            self._keep_alive = None

        self._close_code = code
        try:
            await self.socket.close(code, reason)
        except Exception:
            _log.debug('Ignoring exception closing Gateway socket.', exc_info=True)


DVWS = TypeVar('DVWS', bound='DiscordVoiceWebSocket')


class DiscordVoiceWebSocket:
    """Implements the websocket protocol for handling voice connections."""

    if TYPE_CHECKING:
        thread_id: int
        _connection: VoiceConnectionState
        gateway: str
        _max_heartbeat_timeout: float

    # fmt: off
    IDENTIFY                       = 0
    SELECT_PROTOCOL                = 1
    READY                          = 2
    HEARTBEAT                      = 3
    SESSION_DESCRIPTION            = 4
    SPEAKING                       = 5
    HEARTBEAT_ACK                  = 6
    RESUME                         = 7
    HELLO                          = 8
    RESUMED                        = 9
    CLIENTS_CONNECT                = 11
    VIDEO                          = 12
    CLIENT_DISCONNECT              = 13
    VOICE_BACKEND_VERSION          = 16
    DAVE_PREPARE_TRANSITION        = 21
    DAVE_EXECUTE_TRANSITION        = 22
    DAVE_TRANSITION_READY          = 23
    DAVE_PREPARE_EPOCH             = 24
    MLS_EXTERNAL_SENDER            = 25
    MLS_KEY_PACKAGE                = 26
    MLS_PROPOSALS                  = 27
    MLS_COMMIT_WELCOME             = 28
    MLS_ANNOUNCE_COMMIT_TRANSITION = 29
    MLS_WELCOME                    = 30
    MLS_INVALID_COMMIT_WELCOME     = 31
    # fmt: on

    def __init__(
        self,
        socket: AsyncWebSocket,
        loop: asyncio.AbstractEventLoop,
        *,
        hook: Optional[Callable[..., Coroutine[Any, Any, None]]] = None,
    ) -> None:
        self.ws: AsyncWebSocket = socket
        self.loop: asyncio.AbstractEventLoop = loop
        self._keep_alive: Optional[VoiceKeepAliveHandler] = None
        self._close_code: Optional[int] = None
        self.secret_key: Optional[List[int]] = None
        self.seq_ack: int = -1
        self.voice_version: Optional[str] = None
        self.rtc_worker_version: Optional[str] = None
        if hook:
            self._hook = hook  # type: ignore

    async def _hook(self, *args: Any) -> None:
        pass

    async def _sendstr(self, data: str, /) -> None:
        try:
            await self.ws.send_str(data)
        except WebSocketError:
            if self.ws.closed:
                # Not much we can do here
                _log.debug('Voice socket is closed, cannot send data.')
            else:
                raise

    async def send_as_json(self, data: Any) -> None:
        _log.debug('Voice socket sending: %s.', data)
        await self._sendstr(utils._to_json(data))

    async def send_binary(self, opcode: int, data: bytes) -> None:
        _log.debug('Voice socket sending binary: opcode=%s, size=%d.', opcode, len(data))
        await self.ws.send_bytes(bytes([opcode]) + data)

    send_heartbeat = send_as_json

    async def resume(self) -> None:
        state = self._connection
        payload = {
            'op': self.RESUME,
            'd': {
                'token': state.token,
                'server_id': str(state.server_id),
                'session_id': state.session_id,
                'seq_ack': self.seq_ack,
            },
        }
        await self.send_as_json(payload)

    async def identify(self) -> None:
        state = self._connection
        payload = {
            'op': self.IDENTIFY,
            'd': {
                'server_id': str(state.server_id),
                'user_id': str(state.user.id),
                'session_id': state.session_id,
                'token': state.token,
                'max_dave_protocol_version': state.max_dave_protocol_version,
            },
        }
        await self.send_as_json(payload)

    @classmethod
    async def from_connection_state(
        cls,
        state: VoiceConnectionState,
        *,
        resume: bool = False,
        hook: Optional[Callable[..., Coroutine[Any, Any, Any]]] = None,
        seq_ack: int = -1,
    ) -> Self:
        """Creates a voice websocket for the :class:`VoiceClient`."""
        gateway = f'wss://{state.endpoint}/?v=8'
        client = state.voice_client
        http = client._state.http
        socket = await http.ws_connect(gateway)
        ws = cls(socket, loop=client.loop, hook=hook)
        ws.gateway = gateway
        ws.seq_ack = seq_ack
        ws._connection = state
        ws._max_heartbeat_timeout = 60.0
        ws.thread_id = threading.get_ident()

        if resume:
            await ws.resume()
        else:
            await ws.identify()

        return ws

    async def select_protocol(self, ip: str, port: int, mode: str) -> None:
        payload = {
            'op': self.SELECT_PROTOCOL,
            'd': {
                'protocol': 'udp',
                'data': {
                    'address': ip,
                    'port': port,
                    'mode': mode,
                },
            },
        }

        await self.send_as_json(payload)

    async def client_connect(self) -> None:
        payload = {
            'op': self.VIDEO,
            'd': {
                'audio_ssrc': self._connection.ssrc,
            },
        }

        await self.send_as_json(payload)

    async def speak(self, state: SpeakingState = SpeakingState.voice) -> None:
        payload = {
            'op': self.SPEAKING,
            'd': {
                'speaking': int(state),
                'delay': 0,
                'ssrc': self._connection.ssrc,
            },
        }

        await self.send_as_json(payload)

    async def request_voice_backend_version(self) -> None:
        payload = {
            'op': self.VOICE_BACKEND_VERSION,
            'd': {},
        }

        await self.send_as_json(payload)

    async def send_transition_ready(self, transition_id: int):
        payload = {
            'op': DiscordVoiceWebSocket.DAVE_TRANSITION_READY,
            'd': {
                'transition_id': transition_id,
            },
        }

        await self.send_as_json(payload)

    async def received_message(self, msg: Dict[str, Any]) -> None:
        _log.debug('Voice socket event: %s.', msg)
        op = msg['op']
        data = msg['d']  # According to Discord this key is always given
        self.seq_ack = msg.get('seq', self.seq_ack)

        if op == self.READY:
            await self.initial_connection(data)
        elif op == self.HEARTBEAT_ACK:
            if self._keep_alive:
                self._keep_alive.ack()
        elif op == self.RESUMED:
            _log.debug('Voice RESUME succeeded.')
        elif op == self.SESSION_DESCRIPTION:
            self._connection.mode = data['mode']
            await self.load_secret_key(data)
            self._connection.dave_protocol_version = data['dave_protocol_version']
            if data['dave_protocol_version'] > 0:
                await self._connection.reinit_dave_session()
        elif op == self.HELLO:
            interval = data['heartbeat_interval'] / 1000.0
            self._keep_alive = VoiceKeepAliveHandler(ws=self, interval=interval)
            self._keep_alive.start()
        elif op == self.VOICE_BACKEND_VERSION:
            self.voice_version = data.get('voice')
            self.rtc_worker_version = data.get('rtc_worker')
            _log.debug('Voice backend version: voice=%r, rtc_worker=%r.', self.voice_version, self.rtc_worker_version)
        elif self._connection.dave_session:
            state = self._connection
            if op == self.DAVE_PREPARE_TRANSITION:
                _log.debug(
                    'Preparing for DAVE transition ID %d for protocol version %d.',
                    data['transition_id'],
                    data['protocol_version'],
                )
                state.dave_pending_transitions[data['transition_id']] = data['protocol_version']
                if data['transition_id'] == 0:
                    await state._execute_transition(data['transition_id'])
                else:
                    if data['protocol_version'] == 0 and state.dave_session:
                        state.dave_session.set_passthrough_mode(True, 120)

                    await self.send_transition_ready(data['transition_id'])
            elif op == self.DAVE_EXECUTE_TRANSITION:
                _log.debug('Executing DAVE transition ID %d.', data['transition_id'])
                await state._execute_transition(data['transition_id'])
            elif op == self.DAVE_PREPARE_EPOCH:
                _log.debug('Preparing for DAVE epoch %d.', data['epoch'])
                # When the epoch ID is equal to 1, this message indicates that a new MLS group is to be created for the given protocol version.
                if data['epoch'] == 1:
                    state.dave_protocol_version = data['protocol_version']
                    await state.reinit_dave_session()

        await self._hook(self, msg)

    async def received_binary_message(self, msg: bytes) -> None:
        self.seq_ack = struct.unpack_from('>H', msg, 0)[0]
        op = msg[2]
        _log.debug('Voice socket binary frame: %d bytes, seq=%s, op=%s.', len(msg), self.seq_ack, op)
        state = self._connection

        if state.dave_session is None:
            return

        if op == self.MLS_EXTERNAL_SENDER:
            state.dave_session.set_external_sender(msg[3:])
            _log.debug('Set MLS external sender.')
        elif op == self.MLS_PROPOSALS:
            optype = msg[3]
            result = state.dave_session.process_proposals(
                davey.ProposalsOperationType.append if optype == 0 else davey.ProposalsOperationType.revoke, msg[4:]
            )
            if isinstance(result, davey.CommitWelcome):
                await self.send_binary(
                    DiscordVoiceWebSocket.MLS_COMMIT_WELCOME,
                    result.commit + result.welcome if result.welcome else result.commit,
                )
            _log.debug('MLS proposals processed.')
        elif op == self.MLS_ANNOUNCE_COMMIT_TRANSITION:
            transition_id = struct.unpack_from('>H', msg, 3)[0]
            try:
                state.dave_session.process_commit(msg[5:])
                if transition_id != 0:
                    state.dave_pending_transitions[transition_id] = state.dave_protocol_version
                    await self.send_transition_ready(transition_id)
                _log.debug('MLS commit processed for transition ID %d.', transition_id)
            except Exception:
                await state._recover_from_invalid_commit(transition_id)
        elif op == self.MLS_WELCOME:
            transition_id = struct.unpack_from('>H', msg, 3)[0]
            try:
                state.dave_session.process_welcome(msg[5:])
                if transition_id != 0:
                    state.dave_pending_transitions[transition_id] = state.dave_protocol_version
                    await self.send_transition_ready(transition_id)
                _log.debug('MLS welcome processed for transition ID %d.', transition_id)
            except Exception:
                await state._recover_from_invalid_commit(transition_id)

    async def initial_connection(self, data: Dict[str, Any]) -> None:
        state = self._connection
        state.ssrc = data['ssrc']
        state.voice_port = data['port']
        state.endpoint_ip = data['ip']

        await self.request_voice_backend_version()
        _log.debug('Connecting to voice socket...')
        await self.loop.sock_connect(state.socket, (state.endpoint_ip, state.voice_port))

        state.ip, state.port = await self.discover_ip()
        modes = [mode for mode in data['modes'] if mode in self._connection.supported_modes]
        _log.debug('Received supported encryption modes: %s.', ', '.join(modes))

        mode = modes[0]
        await self.select_protocol(state.ip, state.port, mode)
        _log.debug('Selected the voice protocol for use: %s.', mode)

    async def discover_ip(self) -> Tuple[str, int]:
        state = self._connection
        packet = bytearray(74)
        struct.pack_into('>H', packet, 0, 1)  # 1 = Send
        struct.pack_into('>H', packet, 2, 70)  # 70 = Length
        struct.pack_into('>I', packet, 4, state.ssrc)

        _log.debug('Sending IP discovery packet...')
        await self.loop.sock_sendall(state.socket, packet)

        fut: asyncio.Future[bytes] = self.loop.create_future()

        def get_ip_packet(data: bytes):
            if data[1] == 0x02 and len(data) == 74:
                self.loop.call_soon_threadsafe(fut.set_result, data)

        fut.add_done_callback(lambda f: state.remove_socket_listener(get_ip_packet))
        state.add_socket_listener(get_ip_packet)
        recv = await fut

        _log.debug('Received IP discovery packet: %s.', recv)

        # The IP is ascii starting at the 8th byte and ending at the first null
        ip_start = 8
        ip_end = recv.index(0, ip_start)
        ip = recv[ip_start:ip_end].decode('ascii')

        port = struct.unpack_from('>H', recv, len(recv) - 2)[0]
        _log.debug('Detected IP: %s, port: %s.', ip, port)

        return ip, port

    @property
    def latency(self) -> float:
        """:class:`float`: Latency between a HEARTBEAT and its HEARTBEAT_ACK in seconds."""
        heartbeat = self._keep_alive
        return float('inf') if heartbeat is None else heartbeat.latency

    @property
    def average_latency(self) -> float:
        """:class:`float`: Average of last 20 HEARTBEAT latencies."""
        heartbeat = self._keep_alive
        if heartbeat is None or not heartbeat.recent_ack_latencies:
            return float('inf')

        return sum(heartbeat.recent_ack_latencies) / len(heartbeat.recent_ack_latencies)

    async def load_secret_key(self, data: Dict[str, Any]) -> None:
        _log.debug('Received secret key for voice connection.')
        self.secret_key = self._connection.secret_key = data['secret_key']

        # Send a speak command with the "not speaking" state
        # This also tells Discord our SSRC value, which Discord requires before
        # sending any voice data (and is the real reason why we call this here)
        await self.speak(SpeakingState.none)

    async def poll_event(self) -> None:
        # This exception is handled up the chain
        msg, flags = await asyncio.wait_for(self.ws.recv(), timeout=self._max_heartbeat_timeout)
        if msg is None:
            # Should never happen
            return

        if flags & CurlWsFlag.TEXT:
            await self.received_message(utils._from_json(msg))
        elif flags & CurlWsFlag.BINARY:
            await self.received_binary_message(msg)
        elif flags & CurlWsFlag.CLOSE:
            socket = self.ws
            _log.info(f'Voice socket received close code {socket.close_code} and reason {socket.close_reason!r}.')
            raise ConnectionClosed(socket.close_code or self._close_code, socket.close_reason or '')

    async def close(self, code: int = 1000, reason: bytes = b'') -> None:
        if self._keep_alive:
            self._keep_alive.stop()

        self._close_code = code
        try:
            await self.ws.close(code, reason)
        except Exception:
            _log.debug('Ignoring exception closing voice socket.', exc_info=True)



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from datetime import date
from operator import attrgetter
from typing import TYPE_CHECKING, Optional, Union

from .channel import PartialMessageable
from .enums import ReadStateType, try_enum
from .flags import ReadStateFlags
from .threads import Thread
from .utils import DISCORD_EPOCH, MISSING, parse_time

if TYPE_CHECKING:
    from datetime import datetime

    from typing_extensions import Self

    from .abc import MessageableChannel
    from .guild import Guild
    from .state import ConnectionState
    from .types.read_state import ReadState as ReadStatePayload
    from .user import ClientUser

# fmt: off
__all__ = (
    'ReadState',
)
# fmt: on


class ReadState:
    """Represents the read state of a resource.

    This is a purposefuly very low-level object.

    .. container:: operations

        .. describe:: x == y

            Checks if two read states are equal.

        .. describe:: x != y

            Checks if two read states are not equal.

        .. describe:: hash(x)

            Returns the read state's hash.

    .. versionadded:: 2.1

    Attributes
    -----------
    id: :class:`int`
        The ID of the resource.
    type: :class:`ReadStateType`
        The type of the read state.
    last_acked_id: :class:`int`
        The ID of the last acknowledged entity (e.g. message) in the read state.
        It may *not* point to an existing or valid entity.
    acked_pin_timestamp: Optional[:class:`datetime.datetime`]
        When the channel's pins were last acknowledged.
    badge_count: :class:`int`
        The number of badges in the read state (e.g. mentions).
    last_viewed: Optional[:class:`datetime.date`]
        When the resource was last viewed. Only tracked for read states of type :attr:`ReadStateType.channel`.
    """

    __slots__ = (
        'id',
        'type',
        'last_acked_id',
        'acked_pin_timestamp',
        'badge_count',
        'last_viewed',
        '_flags',
        '_last_entity_id',
        '_state',
    )

    def __init__(self, *, state: ConnectionState, data: ReadStatePayload):
        self._state = state

        self.id: int = int(data['id'])
        self.type: ReadStateType = try_enum(ReadStateType, data.get('read_state_type', 0))
        self._last_entity_id: Optional[int] = None
        self._flags: int = 0
        self.last_viewed: Optional[date] = self.unpack_last_viewed(0) if self.type == ReadStateType.channel else None
        self._update(data)

    def _update(self, data: ReadStatePayload):
        self.last_acked_id: int = int(data.get('last_acked_id', data.get('last_message_id', 0)))
        self.acked_pin_timestamp: Optional[datetime] = parse_time(data.get('last_pin_timestamp'))
        self.badge_count: int = int(data.get('badge_count', data.get('mention_count', 0)))
        if 'flags' in data and data['flags'] is not None:
            self._flags = data['flags']
        if 'last_viewed' in data and data['last_viewed']:
            self.last_viewed = self.unpack_last_viewed(data['last_viewed'])

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ReadState):
            return other.id == self.id and other.type == self.type
        return False

    def __ne__(self, other: object) -> bool:
        if isinstance(other, ReadState):
            return other.id != self.id or other.type != self.type
        return True

    def __hash__(self) -> int:
        return (self.id * self.type.value) >> 22

    @classmethod
    def default(cls, id: int, type: ReadStateType, *, state: ConnectionState) -> Self:
        self = cls.__new__(cls)
        self._state = state
        self.id = id
        self.type = type
        self._last_entity_id = None
        self._flags = 0
        self.last_viewed = cls.unpack_last_viewed(0) if type == ReadStateType.channel else None
        self.last_acked_id = 0
        self.acked_pin_timestamp = None
        self.badge_count = 0
        return self

    @staticmethod
    def unpack_last_viewed(last_viewed: int) -> date:
        # last_viewed is days since the Discord epoch
        return date.fromtimestamp(DISCORD_EPOCH / 1000 + last_viewed * 86400)

    @staticmethod
    def pack_last_viewed(last_viewed: date) -> int:
        # We always round up
        return int((last_viewed - date.fromtimestamp(DISCORD_EPOCH / 1000)).total_seconds() / 86400 + 0.5)

    @property
    def flags(self) -> ReadStateFlags:
        """:class:`ReadStateFlags`: The read state's flags."""
        return ReadStateFlags._from_value(self._flags)

    @property
    def resource(self) -> Union[ClientUser, Guild, MessageableChannel]:
        """Union[:class:`ClientUser`, :class:`Guild`, :class:`TextChannel`, :class:`StageChannel`, :class:`VoiceChannel`, :class:`Thread`, :class:`DMChannel`, :class:`GroupChannel`, :class:`PartialMessageable`]: The entity associated with the read state."""
        state = self._state

        if self.type == ReadStateType.channel:
            return state._get_or_create_partial_messageable(self.id)  # type: ignore
        elif self.type in (ReadStateType.scheduled_events, ReadStateType.guild_home, ReadStateType.onboarding):
            return state._get_or_create_unavailable_guild(self.id)
        elif self.type in (ReadStateType.notification_center, ReadStateType.message_requests) and self.id == state.self_id:
            return state.user  # type: ignore
        else:
            raise NotImplementedError(f'Unknown read state type {self.type!r}')

    @property
    def last_entity_id(self) -> int:
        """:class:`int`: The ID of the last resource (e.g. message) in the read state.
        It may *not* point to an existing or valid resource.
        """
        if self._last_entity_id is not None:
            return self._last_entity_id
        resource = self.resource
        if not resource:
            return 0

        if self.type == ReadStateType.channel:
            return resource.last_message_id or 0  # type: ignore
        elif self.type == ReadStateType.scheduled_events:
            return max(resource.scheduled_events, key=attrgetter('id')).id  # type: ignore
        return 0

    @property
    def last_pin_timestamp(self) -> Optional[datetime]:
        """Optional[:class:`datetime.datetime`]: When the last pinned message was pinned in the channel."""
        if self.resource and hasattr(self.resource, 'last_pin_timestamp'):
            return self.resource.last_pin_timestamp  # type: ignore

    async def ack(
        self,
        entity_id: int,
        *,
        manual: bool = False,
        mention_count: Optional[int] = None,
        last_viewed: Optional[date] = MISSING,
    ) -> None:
        """|coro|

        Updates the read state. This is a purposefully low-level function.

        Parameters
        -----------
        entity_id: :class:`int`
            The ID of the entity to set the read state to.
        manual: :class:`bool`
            Whether the read state was manually set by the user.
            Only for read states of type :attr:`ReadStateType.channel`.
        mention_count: Optional[:class:`int`]
            The number of mentions to set the read state to. Only applicable for
            manual acknowledgements. Only for read states of type :attr:`ReadStateType.channel`.
        last_viewed: Optional[:class:`datetime.date`]
            The last day the user viewed the channel. Defaults to today for non-manual acknowledgements.
            Only for read states of type :attr:`ReadStateType.channel`.

        Raises
        -------
        ValueError
            Invalid parameters were passed.
        HTTPException
            Updating the read state failed.
        """
        state = self._state

        if self.type == ReadStateType.channel:
            flags = None
            channel: MessageableChannel = self.resource  # type: ignore
            if not isinstance(channel, PartialMessageable):
                # Read state flags are kept accurate by the client 😭
                flags = ReadStateFlags()
                if isinstance(channel, Thread):
                    flags.thread = True
                elif channel.guild:
                    flags.guild_channel = True

                if flags == self.flags:
                    flags = None

            if not manual and last_viewed is MISSING:
                last_viewed = date.today()

            await state.http.ack_message(
                self.id,
                entity_id,
                manual=manual,
                mention_count=mention_count,
                flags=flags.value if flags else None,
                last_viewed=self.pack_last_viewed(last_viewed) if last_viewed else None,
            )
            return

        if manual or mention_count is not None or last_viewed:
            raise ValueError('Extended read state parameters are only valid for channel read states')

        if self.type in (ReadStateType.scheduled_events, ReadStateType.guild_home, ReadStateType.onboarding):
            await state.http.ack_guild_feature(self.id, self.type.value, entity_id)
        elif self.type == ReadStateType.notification_center:
            await state.http.ack_user_feature(self.type.value, entity_id)

    async def delete(self):
        """|coro|

        Deletes the read state.

        Raises
        -------
        HTTPException
            Deleting the read state failed.
        """
        state = self._state
        await state.http.delete_read_state(self.id, self.type.value)
        state.remove_read_state(self)
